from importlib.util import find_spec
from functools import wraps
from math import isclose
from pathlib import Path
from typing import Collection, Any, Type, List, TypeVar, Iterator

# Conditional imports
if find_spec('xlrd'):
    import xlrd
    from xlrd.sheet import Sheet
else:
    xlrd = None
    Sheet = TypeVar('Sheet')

if find_spec('xlwt'):
    import xlwt
    from xlwt import Workbook as WriteWorkbook, Worksheet as WriteWorksheet
else:
    xlwt = None
    WriteWorkbook = TypeVar('WriteWorkbook')
    WriteWorksheet = TypeVar('WriteWorksheet')

if find_spec('xlutils'):
    import xlutils
    from xlutils.copy import copy as xlutils_copy
else:
    xlutils = None

from ..core import requires
from ..spreadsheets import SpreadsheetSerializer
from ..typing import ParsedDataRow


__ALL__ = ['XLSSerializer', 'dump', 'write_table', 'load', 'read_table', 'read_headers', 'read_schema', 'iter_table',
           'iter_structure']


class XLSSerializer(SpreadsheetSerializer):
    MAX_LINES = 65536
    MAX_COLS = 256

    def __init__(self, int_precision: float | None = 1E-9, **kwargs):
        super().__init__(**kwargs)
        self.int_precision = int_precision

    #  xlwt specific implementation
    @staticmethod
    @requires('xlwt')
    def _add_sheet(workbook: WriteWorkbook, sheet_name: str) -> WriteWorksheet:
        return workbook.add_sheet(sheet_name)

    @staticmethod
    @requires('xlwt')
    def _write_headers(sheet: WriteWorksheet, headers: Collection[str]):
        for header_col, header_title in enumerate(headers):
            sheet.write(0, header_col, header_title)

    @staticmethod
    @requires('xlwt')
    def _write_row(sheet: WriteWorksheet, headers: Collection[str], row_num: int, row_data: ParsedDataRow,
                   default: Any = SpreadsheetSerializer.DEFAULT_EMPTY_ON_WRITE):
        for col_num, col_key in enumerate(headers):
            sheet.write(row_num, col_num, row_data.get(col_key, default))

    @staticmethod
    @requires('xlwt')
    def _freeze_headers(sheet: WriteWorksheet):
        sheet.set_panes_frozen(True)  # frozen headings instead of split panes
        sheet.set_horz_split_pos(1)  # in general, freeze after last heading row
        sheet.set_remove_splits(True)  # if user does unfreeze, don't leave a split there

    @staticmethod
    @requires('xlwt')
    def _get_new_workbook():
        return xlwt.Workbook()

    @staticmethod
    @requires('xlwt')
    def _save_workbook(workbook, fp: Path | str):
        workbook.save(fp)

    # xlrd specific implementation
    @staticmethod
    @requires('xlrd')
    def _get_sheet_name(sheet) -> str:
        return sheet.name

    @staticmethod
    @requires('xlrd')
    def _get_all_sheets(book) -> List[Any]:
        return book.sheets()

    @requires('xlrd')
    def _iter_through_sheet(self,
                            sheet: Sheet,
                            workbook,
                            constructor: Type = None,
                            default: Any = SpreadsheetSerializer.DEFAULT_EMPTY_ON_READ,
                            ) -> Iterator[ParsedDataRow]:

        has_formatting_info = getattr(workbook, 'has_formatting_info', False)
        int_precision = self.int_precision

        # Extract headers from first row
        headers = list()
        for col_num in range(sheet.ncols):
            headers.append(sheet.cell_value(0, col_num))

        # Extract data from remaining rows
        for row_num in range(1, sheet.nrows):

            row = dict()

            for col_num in range(sheet.ncols):
                value = sheet.cell_value(row_num, col_num)
                value = value if value != '' else default

                if has_formatting_info:
                    value = self._format_cell(value, sheet.cell_type(row_num, col_num), workbook.datemode, default)

                # Convert floats to ints if they are close to an integer
                if int_precision and isinstance(value, float) and isclose(value, int(value), abs_tol=int_precision):
                    value = int(value)

                row[headers[col_num]] = value

            yield constructor(**row) if constructor else row

    @requires('xlrd')
    def _format_cell(self, value, xl_index, date_mode,
                     default: Any = SpreadsheetSerializer.DEFAULT_EMPTY_ON_READ) -> Any:
        match xl_index:
            case xlrd.XL_CELL_EMPTY:
                return default
            case xlrd.XL_CELL_TEXT:
                return str(value)
            case xlrd.XL_CELL_NUMBER:
                return float(value)
            case xlrd.XL_CELL_DATE:
                return xlrd.xldate_as_tuple(value, date_mode)
            case xlrd.XL_CELL_BOOLEAN:
                return bool(value)
            case xlrd.XL_CELL_ERROR:
                return self.CellError.match_behaviour(self.cell_error, value, default)
            case xlrd.XL_CELL_BLANK:
                return default
            case _:
                return value

    @requires('xlrd')
    def _open_workbook(self, fp: Path | str, for_append=False, **kwargs) -> 'xlrd.Workbook':
        try:
            has_formatting_info = kwargs.setdefault('formatting_info', True)
            workbook = xlrd.open_workbook(Path(fp), **kwargs)
        except NotImplementedError:
            # Catches formatting_info not implemented for 2007 Excel files
            has_formatting_info = False
            kwargs['formatting_info'] = False
            workbook = xlrd.open_workbook(Path(fp), **kwargs)

        workbook.has_formatting_info = has_formatting_info

        return self._convert_for_append(workbook) if for_append else workbook

    @staticmethod
    @requires('xlutils')
    def _convert_for_append(workbook):
        return xlutils_copy(workbook)


# Default implementation
__xls_serializer = XLSSerializer()

write_table = wraps(__xls_serializer.write_table)
dump = wraps(__xls_serializer.dump)
load = wraps(__xls_serializer.load)
read_table = wraps(__xls_serializer.read_table)
read_headers = wraps(__xls_serializer.read_headers)
read_schema = wraps(__xls_serializer.read_schema)
iter_table = wraps(__xls_serializer.iter_table)
iter_structure = wraps(__xls_serializer.iter_structure)



from abc import abstractmethod
from collections import defaultdict
from collections.abc import Mapping
from dataclasses import asdict, is_dataclass
from enum import IntEnum
from fnmatch import fnmatchcase
from operator import itemgetter
from pathlib import Path
from typing import Collection, Any, Iterable, Type, Tuple, Iterator, Callable, Dict, List, TypeVar

from .core import Serializer
from .typing import DataTable, DataStructure, DataRow, ParsedDataRow, ParsedDataStructure, ParsedDataTable


Workbook = TypeVar('Workbook')
Worksheet = TypeVar('Worksheet')
Cell = TypeVar('Cell')


class SpreadsheetSerializer(Serializer):
    DEFAULT_NAMING_PATTERN = f'{{table_name}}_{{table_series}}'
    DEFAULT_LOAD_PATTERN = '*'
    DEFAULT_EMPTY_ON_READ = None
    DEFAULT_EMPTY_ON_WRITE = ''
    DEFAULT_TABLE_NAME = 'data'

    MAX_LINES = NotImplemented
    MAX_COLS = NotImplemented

    class CellError:
        class Behaviour(IntEnum):
            IGNORE = 0
            WRAP = 1
            RAISE = 255

        class CellErrorWrapper:
            def __init__(self, value: Any):
                self.value = value

        @classmethod
        def match_behaviour(cls,
                            behaviour: Behaviour,
                            value: Any,
                            default: Any = None,
                            wrapper: Callable = CellErrorWrapper,
                            ) -> Any:
            match behaviour:
                case cls.Behaviour.IGNORE:
                    return default
                case cls.Behaviour.WRAP:
                    return wrapper(value)
                case cls.Behaviour.RAISE:
                    raise ValueError(f'cell error found: {value}')
                case _:
                    raise ValueError(f'unknown behaviour: {behaviour}')

    def __init__(self, cell_error: CellError.Behaviour = CellError.Behaviour.WRAP):
        super().__init__()
        self.cell_error = cell_error

    # Generic XLS/XLSX implementation
    def write_table(self, table: DataTable, fp: Path | str, table_name: str = DEFAULT_TABLE_NAME,
                    headers: Collection[str] = None, default: Any = DEFAULT_EMPTY_ON_WRITE, **serializer_kwargs):

        workbook = self._get_new_workbook()
        self._add_table(workbook, table_name=table_name, table=table, default=default, headers=headers)
        self._save_workbook(workbook, fp)

    def append_table(self, table: DataTable, fp: Path | str, table_name: str = DEFAULT_TABLE_NAME,
                     headers: Collection[str] = None, default: Any = DEFAULT_EMPTY_ON_WRITE, **serializer_kwargs):

        workbook = self._open_workbook(fp, for_append=True)
        self._add_table(workbook, table_name, table, default=default, headers=headers)
        self._save_workbook(workbook, fp)

    def dump(self, structure: DataStructure, fp: Path | str, schema: Mapping[Collection[str]] = None,
             default: Any = DEFAULT_EMPTY_ON_WRITE, **serializer_kwargs) -> None:
        schema = schema or dict()
        if isinstance(structure, Mapping):
            structure = structure.items()

        workbook = self._get_new_workbook()
        for table_name, table in structure:
            if schema and table_name not in schema:
                continue

            self._add_table(workbook, table_name, table, default=default, headers=schema.get(table_name, None))
        self._save_workbook(workbook, fp)

    def load(self, fp: Path | str, pattern=DEFAULT_LOAD_PATTERN, constructor: Type = None,
             default: Any = DEFAULT_EMPTY_ON_READ, **open_kwargs) -> ParsedDataStructure:

        structure = defaultdict(list)
        for table_name, row in self.iter_structure(fp, pattern=pattern, constructor=constructor,
                                                   default=default, **open_kwargs):
            structure[table_name].append(row)
        return structure

    def read_table(self, fp: Path | str, table_name: str = DEFAULT_TABLE_NAME, constructor: Type = None,
                   default: Any = DEFAULT_EMPTY_ON_READ, **open_kwargs) -> ParsedDataTable:
        return list(self.iter_table(fp, table_name=table_name, constructor=constructor,
                                    default=default, **open_kwargs))

    def read_headers(self, fp: Path | str, table_name: str = DEFAULT_TABLE_NAME, constructor: Type = None,
                     default: Any = DEFAULT_EMPTY_ON_READ, **open_kwargs) -> ParsedDataTable:
        first_row = next(self.iter_table(fp, table_name=table_name, constructor=constructor,
                                         default=default, **open_kwargs))
        return first_row.keys()

    def read_schema(self, fp: Path | str, pattern=DEFAULT_LOAD_PATTERN, **open_kwargs) -> Mapping[Collection[str]]:
        workbook = self._open_workbook(fp, **open_kwargs)
        return dict(self._iter_schema(workbook, pattern=pattern))

    def iter_table(self, fp: Path | str, table_name: str = None, constructor: Type = None,
                   default: Any = DEFAULT_EMPTY_ON_READ, **open_kwargs) -> Iterator[ParsedDataRow]:
        workbook = self._open_workbook(fp, **open_kwargs)
        table_name = table_name or self.DEFAULT_TABLE_NAME
        yield from self._iter_table(table_name, workbook, constructor, default)

    def iter_structure(self, fp: Path | str, pattern=DEFAULT_LOAD_PATTERN, constructor: Type = None,
                       default: Any = DEFAULT_EMPTY_ON_READ, **open_kwargs) -> Iterator[Tuple[str, ParsedDataRow]]:
        workbook = self._open_workbook(fp, **open_kwargs)
        yield from self._iter_structure(workbook, pattern, constructor, default)

    @staticmethod
    def _parse_table_name(s: str) -> Tuple[str, int | None]:
        table_name, *table_series = s.rsplit('_', 1)
        try:
            return table_name, int(table_series[0])
        except (IndexError, ValueError):
            return s, None

    def _add_table(self,
                   workbook,
                   table_name: str,
                   table: Iterable[DataRow],
                   default: Any = DEFAULT_EMPTY_ON_WRITE,
                   headers: Collection[str] = None,
                   ):
        max_lines = self.MAX_LINES
        max_cols = self.MAX_COLS
        naming_pattern = self.DEFAULT_NAMING_PATTERN

        table_series = 1
        sheet = None
        top_row = 0

        for row_num, row in enumerate(table):
            row = asdict(row) if is_dataclass(row) else row

            # Capture headers if None, presumes data is normalized though
            if not headers:
                headers = row.keys()

                if len(headers) > max_cols:
                    raise ValueError(f'too many columns for XLS: {len(headers)} > {max_cols} (max supported)')

            # handles new sheet
            if not sheet or (row_num - top_row + 1) >= max_lines:
                sheet_name = naming_pattern.format(table_name=table_name, table_series=table_series)
                sheet = self._add_sheet(workbook, sheet_name)

                self._write_headers(sheet, headers)
                self._freeze_headers(sheet)

                table_series += 1
                top_row = row_num

            # write data
            self._write_row(sheet, headers, row_num - top_row + 1, row, default=default)

    def _find_tables(self, workbook, pattern='*') -> Dict[str, List[Any]]:
        tables = defaultdict(list)

        for sheet in self._get_all_sheets(workbook):
            if pattern == '*' or fnmatchcase(sheet.name, pattern):
                sheet_name = self._get_sheet_name(sheet)
                table_name, table_series = self._parse_table_name(sheet_name)
                tables[table_name].append(((-1 if table_series is None else table_series), sheet))

        # Sort and clean-up serials
        for table_name, table_sheets in tables.items():
            table_sheets.sort(key=itemgetter(0))
            tables[table_name] = [sheet for _, sheet in table_sheets]

        return tables

    def _find_table_sheets(self, table_name, workbook) -> List[Worksheet]:
        table_sheets = list()

        for sheet in self._get_all_sheets(workbook):
            this_name, this_series = self._parse_table_name(self._get_sheet_name(sheet))
            if table_name == this_name:
                table_sheets.append(((-1 if this_series is None else this_series), sheet))

        table_sheets.sort(key=itemgetter(0))
        return [sheet for _, sheet in table_sheets]

    def _iter_structure(self,
                        workbook,
                        pattern: str = '*',
                        constructor: Type = None,
                        default: Any = DEFAULT_EMPTY_ON_READ,
                        ) -> Iterator[Tuple[str, Iterable[DataRow]]]:

        tables = self._find_tables(workbook, pattern=pattern)

        for table_name, table_sheets in tables.items():
            for table_sheet in table_sheets:
                for row in self._iter_through_sheet(table_sheet, workbook, constructor, default):
                    yield table_name, row

    def _iter_table(self,
                    table_name: str,
                    workbook,
                    constructor: Type = None,
                    default: Any = DEFAULT_EMPTY_ON_READ,
                    ) -> Iterator[Dict[str, Any]]:

        for sheet in self._find_table_sheets(table_name, workbook):
            yield from self._iter_through_sheet(sheet, workbook, constructor, default)

    def _iter_schema(self,
                     workbook: Workbook,
                     pattern: str = '*',
                     ) -> Iterator[Tuple[str, Collection[str]]]:

        tables = self._find_tables(workbook, pattern=pattern)

        for table_name, table_sheets in tables.items():
            for table_sheet in table_sheets:
                first_row = next(self._iter_through_sheet(table_sheet, workbook))
                yield table_name, list(first_row.keys())

    # Abstract methods to be adapted by xls/xlsx implementations
    @abstractmethod
    def _add_sheet(self, workbook: Workbook, sheet_name: str) -> Worksheet:
        pass

    @abstractmethod
    def _write_headers(self, sheet: Worksheet, headers: Collection[str]):
        pass

    @abstractmethod
    def _write_row(self, sheet: Worksheet, headers: Collection[str], row_num: int, row_data: ParsedDataRow,
                   default: Any = DEFAULT_EMPTY_ON_WRITE):
        pass

    @abstractmethod
    def _freeze_headers(self, sheet: Worksheet):
        pass

    @abstractmethod
    def _get_new_workbook(self, **new_kwargs) -> Workbook:
        pass

    @abstractmethod
    def _save_workbook(self, workbook: Workbook, fp: Path | str):
        pass

    @abstractmethod
    def _iter_through_sheet(self,
                            sheet: Worksheet,
                            workbook: Workbook,
                            constructor: Type = None,
                            default: Any = DEFAULT_EMPTY_ON_READ,
                            ) -> Iterator[ParsedDataRow]:
        pass

    @abstractmethod
    def _get_sheet_name(self, sheet: Worksheet) -> str:
        pass

    @abstractmethod
    def _get_all_sheets(self, book: Workbook) -> List[Any]:
        pass

    @abstractmethod
    def _open_workbook(self, fp: Path | str, for_append=False, **kwargs) -> Workbook:
        pass



from collections import defaultdict
from collections.abc import Mapping
from dataclasses import asdict, is_dataclass
from contextlib import ExitStack
from csv import DictWriter, DictReader
from functools import wraps
from pathlib import Path
from typing import Collection, Any, TextIO, Type, Tuple, Iterator

from nv.utils.collections.structures import normalize_row

from ..core import Serializer
from ..typing import DataTable, DataStructure, ParsedDataRow, ParsedDataStructure, ParsedDataTable


__ALL__ = ['CSVSerializer', 'dump', 'write_table', 'load', 'read_table', 'read_headers', 'read_schema', 'iter_table',
           'iter_structure']


class CSVSerializer(Serializer):
    DEFAULT_EXTENSION = 'csv'
    DEFAULT_NAMING_PATTERN = f'{{table_name}}.{DEFAULT_EXTENSION}'
    DEFAULT_LOAD_PATTERN = f'*.{DEFAULT_EXTENSION}'
    DEFAULT_EMPTY_ON_READ = None
    DEFAULT_EMPTY_ON_WRITE = ''
    DEFAULT_ENCODING = 'utf-8'

    def __init__(self, encoding: str = DEFAULT_ENCODING, **fmtparams):
        self.encoding = encoding
        self.fmtparams = fmtparams
        super().__init__()

    def write_table(self, table: DataTable, fp: Path | str | TextIO, headers: Collection[str] = None,
                    default: Any = DEFAULT_EMPTY_ON_WRITE, encoding: str = None, **serializer_kwargs) -> None:

        with ExitStack() as stack:
            if isinstance(fp, str | Path):
                fp = stack.enter_context(Path(fp).open('w', encoding=self._get_encoding(fp, encoding)))

            self._write_table(table, fp, headers, default, **serializer_kwargs)

    def append_table(self, table: DataTable, fp: Path | str | TextIO, headers: Collection[str] = None,
                     default: Any = DEFAULT_EMPTY_ON_WRITE, encoding: str = None, **serializer_kwargs) -> None:

        with ExitStack() as stack:
            if isinstance(fp, str | Path):
                fp = stack.enter_context(Path(fp).open('a', encoding=self._get_encoding(fp, encoding)))

            self._write_table(table, fp, headers, default, append=True, **serializer_kwargs)

    def dump(self, structure: DataStructure, fp: Path | str | TextIO, schema: Mapping[Collection[str]] = None,
             pattern=DEFAULT_NAMING_PATTERN, default: Any = DEFAULT_EMPTY_ON_WRITE, encoding: str = None,
             **serializer_kwargs) -> None:

        schema = schema or dict()

        fp = Path(fp)
        fp.mkdir(parents=True, exist_ok=True)
        if not fp.is_dir():
            raise ValueError(f'{fp} must be a directory')

        if isinstance(structure, Mapping):
            structure = structure.items()

        for table_name, table in structure:
            if schema and table_name not in schema:
                continue

            table_path = fp / Path(pattern.format(table_name=table_name))
            self.write_table(table, table_path, headers=schema.get(table_name, None),
                             default=default, encoding=encoding, **serializer_kwargs)

    def iter_table(self, fp: Path | str | TextIO, constructor: Type = None, default: Any = DEFAULT_EMPTY_ON_READ,
                   encoding: str = None, **parser_kwargs) -> Iterator[ParsedDataRow]:

        with ExitStack() as stack:
            if isinstance(fp, str | Path):
                fp = stack.enter_context(Path(fp).open('r', encoding=self._get_encoding(fp, encoding)))

            dict_reader = DictReader(fp, **(self.fmtparams | parser_kwargs))

            headers = dict_reader.fieldnames

            for row in dict_reader:
                yield constructor(**row) if constructor else normalize_row(row, headers=headers, default=default)

    def iter_structure(self, fp: Path | str, pattern=DEFAULT_LOAD_PATTERN, constructor: Type = None,
                       default: Any = DEFAULT_EMPTY_ON_READ, encoding: str = None,
                       **parser_kwargs) -> Iterator[Tuple[str, ParsedDataRow]]:

        for f in self._iter_tables(fp, pattern):
            for row in self.iter_table(f, constructor=constructor, default=default, encoding=encoding, **parser_kwargs):
                table_name, _ = self._parse_table_name(f)
                yield table_name, row

    def read_table(self, fp: Path | str | TextIO, constructor: Type = None, default: Any = DEFAULT_EMPTY_ON_READ,
                   encoding: str = None, **parser_kwargs) -> ParsedDataTable:
        return list(self.iter_table(fp, constructor=constructor, default=default, encoding=encoding, **parser_kwargs))

    def load(self, fp: Path | str, pattern=DEFAULT_LOAD_PATTERN, constructor: Type = None,
             default: Any = DEFAULT_EMPTY_ON_READ, encoding: str = None, **parser_kwargs) -> ParsedDataStructure:

        structure = defaultdict(list)
        for table_name, row in self.iter_structure(fp, pattern=pattern, constructor=constructor, default=default,
                                                   encoding=encoding, **parser_kwargs):
            structure[table_name].append(row)
        return structure

    def read_headers(self, fp: Path | str | TextIO, encoding: str = None, **parser_kwargs) -> Collection[str]:
        with ExitStack() as stack:
            if isinstance(fp, str | Path):
                fp = stack.enter_context(Path(fp).open('r', encoding=self._get_encoding(fp, encoding)))

            dict_reader = DictReader(fp, **(self.fmtparams | parser_kwargs))
            return dict_reader.fieldnames

    def read_schema(self, fp: Path | str, pattern=DEFAULT_LOAD_PATTERN, encoding: str = None,
                    **parser_kwargs) -> Mapping[Collection[str]]:

        schema = dict()
        for f in self._iter_tables(fp, pattern):
            table_name, _ = self._parse_table_name(f)
            schema[table_name] = self.read_headers(f, encoding=encoding, **parser_kwargs)
        return schema

    @staticmethod
    def _parse_table_name(fp: Path | str) -> Tuple[str, int | None]:
        return fp.stem, None

    def _write_table(self, table: DataTable, fp: TextIO, headers: Collection[str] = None,
                     default: Any = DEFAULT_EMPTY_ON_READ, append=False, **serializer_kwargs) -> None:

        dict_writer = None

        for row in table:
            row = asdict(row) if is_dataclass(row) else row

            # Capture headers if None, presumes data is normalized though
            if not headers:
                headers = (asdict(row) if is_dataclass(row) else row).keys()

            # Write headers
            if dict_writer is None:
                if not headers:
                    # Guards against any empty data
                    raise ValueError('no headers found in data (or no data to write).')

                dict_writer = DictWriter(fp, headers, **(self.fmtparams | serializer_kwargs))

                if not append:
                    dict_writer.writeheader()

            dict_writer.writerow(normalize_row(row, headers=headers, default=default))

    @staticmethod
    def _iter_tables(fp: Path | str, pattern='*.csv') -> Iterator[Path]:
        fp = Path(fp)
        if not fp.is_dir():
            raise ValueError(f'{fp} is not a directory.')

        for f in fp.glob(pattern):
            if f.is_file():
                yield f

    def _get_encoding(self, fp: Path | str, encoding: str = None) -> str:
        encoding = encoding or self.encoding or self.DEFAULT_ENCODING
        if callable(encoding):
            encoding = encoding(fp)
        return encoding


# Default implementation
__csv_serializer = CSVSerializer()

write_table = wraps(__csv_serializer.write_table)
dump = wraps(__csv_serializer.dump)
load = wraps(__csv_serializer.load)
read_table = wraps(__csv_serializer.read_table)
read_headers = wraps(__csv_serializer.read_headers)
read_schema = wraps(__csv_serializer.read_schema)
iter_table = wraps(__csv_serializer.iter_table)
iter_structure = wraps(__csv_serializer.iter_structure)



from importlib.util import find_spec
from functools import wraps
from pathlib import Path
from typing import Collection, Any, Type, List, TypeVar, Iterator

# Conditional imports
if find_spec('openpyxl'):
    import openpyxl
    from openpyxl.cell import WriteOnlyCell
    from openpyxl import Workbook
    from openpyxl.worksheet.worksheet import Worksheet
    from openpyxl.cell.cell import Cell, TYPE_ERROR
else:
    openpyxl = None
    WriteOnlyCell = None
    Cell = None
    TYPE_ERROR = 'e'
    Workbook = TypeVar('Workbook')
    Worksheet = TypeVar('Worksheet')

from nv.utils.collections.structures import normalize_row

from ..core import requires
from ..spreadsheets import SpreadsheetSerializer
from ..typing import ParsedDataRow



__ALL__ = ['XLSXSerializer', 'dump', 'write_table', 'load', 'read_table', 'read_headers', 'read_schema', 'iter_table',
           'iter_structure']


class XLSXSerializer(SpreadsheetSerializer):
    MAX_LINES = 1048576
    MAX_COLS = 16384

    # OpenPyXL specific implementation
    @staticmethod
    @requires('openpyxl')
    def _add_sheet(workbook: Workbook, sheet_name: str) -> Worksheet:
        return workbook.create_sheet(title=sheet_name)

    @staticmethod
    @requires('openpyxl')
    def _write_headers(sheet: Worksheet, headers: Collection[str]):
        sheet.append((WriteOnlyCell(sheet, h) for h in headers))

    @staticmethod
    @requires('openpyxl')
    def _write_row(sheet: Worksheet, headers: Collection[str], row_num: int, row_data: ParsedDataRow,
                   default: Any = SpreadsheetSerializer.DEFAULT_EMPTY_ON_WRITE):
        sheet.append(list(normalize_row(row_data, headers, default=default).values()))

    @staticmethod
    @requires('openpyxl')
    def _freeze_headers(sheet: Worksheet):
        # TODO: This probably requires not using a write-only sheet
        # sheet.freeze_panes = sheet.cell(2, 1)
        pass

    @staticmethod
    @requires('openpyxl')
    def _get_new_workbook(**new_kwargs) -> Workbook:
        new_kwargs.setdefault('write_only', True)
        return openpyxl.Workbook(**new_kwargs)

    @staticmethod
    @requires('openpyxl')
    def _save_workbook(workbook: Workbook, fp: Path | str):
        workbook.save(fp)

    @requires('openpyxl')
    def _iter_through_sheet(self,
                            sheet: Worksheet,
                            workbook: Workbook,
                            constructor: Type = None,
                            default: Any = SpreadsheetSerializer.DEFAULT_EMPTY_ON_READ,
                            ) -> Iterator[ParsedDataRow]:
        headers = None

        for row in sheet.rows:

            # Extract headers from first row
            if headers is None:
                headers = [c.value for c in row]
                continue

            row = dict(zip(headers, (self._parse_cell_value(c, default) for c in row)))

            yield constructor(**row) if constructor is not None else row

    @requires('openpyxl')
    def _parse_cell_value(self, cell: Cell, default: Any = SpreadsheetSerializer.DEFAULT_EMPTY_ON_READ) -> Any:
        value = cell.value
        if cell.data_type == TYPE_ERROR:
            return self.CellError.match_behaviour(self.cell_error, value, default)

        return value

    @staticmethod
    @requires('openpyxl')
    def _get_sheet_name(sheet: Worksheet) -> str:
        return sheet.title

    @staticmethod
    @requires('openpyxl')
    def _get_all_sheets(book: Workbook) -> List[Any]:
        return book.worksheets

    @staticmethod
    @requires('openpyxl')
    def _open_workbook(fp: Path | str, for_append=False, **kwargs) -> Workbook:
        return openpyxl.load_workbook(fp, read_only=not for_append, data_only=not for_append, **kwargs)


# Default implementation
__xlsx_serializer = XLSXSerializer()

write_table = wraps(__xlsx_serializer.write_table)
dump = wraps(__xlsx_serializer.dump)
load = wraps(__xlsx_serializer.load)
read_table = wraps(__xlsx_serializer.read_table)
read_headers = wraps(__xlsx_serializer.read_headers)
read_schema = wraps(__xlsx_serializer.read_schema)
iter_table = wraps(__xlsx_serializer.iter_table)
iter_structure = wraps(__xlsx_serializer.iter_structure)



from typing import Dict, Any, Mapping, List, Iterable, Tuple

from nv.utils.typing import DataClass


# Generic data structures
DataRow = Mapping[str, Any] | DataClass
DataTable = Iterable[DataRow]
DataStructure = Iterable[Tuple[str, DataTable]] | Mapping[str, DataTable]


# Output data structures
ParsedDataRow = Dict[str, Any] | DataClass
ParsedDataTable = List[ParsedDataRow]
ParsedDataStructure = Mapping[str, ParsedDataTable]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['nv', 'nv.tables', 'nv.tables.implementations']

package_data = \
{'': ['*']}

install_requires = \
['nv-utils>=0.3.0a,<0.4.0']

extras_require = \
{'all': ['openpyxl>=3.0.9,<4.0.0',
         'xlrd>=2.0.1,<3.0.0',
         'xlwt>=1.3.0,<2.0.0',
         'xlutils>=2.0.0,<3.0.0'],
 'xls': ['xlrd>=2.0.1,<3.0.0', 'xlwt>=1.3.0,<2.0.0', 'xlutils>=2.0.0,<3.0.0'],
 'xlsx': ['openpyxl>=3.0.9,<4.0.0']}

setup_kwargs = {
    'name': 'nv-tables',
    'version': '0.1.0',
    'description': 'Tools for read and write data from and into table files (e.g. xls, xlsx, csv, etc.)',
    'long_description': "# nv.tables\nTools for parsing data from and into table formats (CSV, XLS and XLSX)\n\n## What's inside?\n[to come]\n\n## Disclaimers\nTHIS IS UNDOCUMENTED WORK IN PROGRESS. READ THE LICENSE AND USE IT AT YOUR OWN RISK.\n\nTHIS IS STILL A BETA AND BREAKING CHANGES MAY (AND PROBABLY WILL) OCCUR UNTIL ITS CONTENT STABILIZES. WE\nARE ACTIVELY MIGRATING STUFF OUT OF THIS LIBRARY (AND LOOKING FOR SUBSTITUTES THAT ARE MORE ACTIVELY MAINTAINED)\n",
    'author': 'Gustavo Santos',
    'author_email': 'gustavo@next.ventures',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/gstos/nv-utils',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



from abc import abstractmethod
from functools import wraps
from importlib.util import find_spec
from pathlib import Path
from typing import Collection, Any, Mapping, Type, Tuple, ParamSpec, TypeVar, Callable, Iterator, Protocol

from .typing import DataTable, DataStructure, ParsedDataRow, ParsedDataTable, ParsedDataStructure


class Serializer:
    @abstractmethod
    def write_table(self, table: DataTable, fp: Path | str, headers: Collection[str] = None, default: Any = None,
                    **serializer_kwargs) -> None:
        raise NotImplementedError

    @abstractmethod
    def append_table(self, table: DataTable, fp: Path | str, headers: Collection[str] = None, default: Any = None,
                     **serializer_kwargs) -> None:
        raise NotImplementedError

    @abstractmethod
    def dump(self, structure: DataStructure, fp: Path | str, schema: Mapping[str, Collection[str]], default: Any = None,
             **serializer_kwargs) -> None:
        raise NotImplementedError

    @abstractmethod
    def iter_table(self, fp: Path | str, constructor: Type = None, default: Any = None,
                   **parser_kwargs) -> Iterator[ParsedDataRow]:
        raise NotImplementedError

    @abstractmethod
    def iter_structure(self, fp: Path | str, constructor: Type = None, default: Any = None,
                       **parser_kwargs) -> Iterator[Tuple[str, ParsedDataRow]]:
        raise NotImplementedError

    @abstractmethod
    def read_table(self, fp: Path | str, constructor: Type = None, default: Any = None,
                   **parser_kwargs) -> ParsedDataTable:
        raise NotImplementedError

    @abstractmethod
    def load(self, fp: Path | str, constructor: Type = None, default: Any = None,
             **parser_kwargs) -> ParsedDataStructure:
        raise NotImplementedError

    @abstractmethod
    def read_headers(self, fp: Path | str, **parser_kwargs) -> Collection[str]:
        pass

    @abstractmethod
    def read_schema(self, fp: Path | str, **parser_kwargs) -> Mapping[str, Collection[str]]:
        pass


P = ParamSpec('P')
V = TypeVar('V')


class DecoratedCallable(Protocol):
    __globals__: Mapping[str, Any]

    @abstractmethod
    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> V:
        pass


def requires(module_name: str, package: str = None, pypi_name: str = None
             ) -> Callable[[DecoratedCallable], DecoratedCallable]:
    """
    requires(module_name: str, package: str = None, pypi_name: str = None
             ) -> Callable[[Callable[[*P.args, *P.kwargs], V]], Callable[[*P.args, *P.kwargs], V]]:
    """

    def decorate(f: DecoratedCallable) -> DecoratedCallable:
        check = find_spec(module_name, package=package) is not None

        @wraps(f)
        def wrapper(*args: P.args, **kwargs: P.kwargs) -> V:
            nonlocal check

            if not check:
                if f.__globals__.get(module_name, None) is None:
                    raise ModuleNotFoundError(f"{pypi_name or module_name} required by {f.__name__ } is not installed")
                else:
                    check = True

            return f(*args, **kwargs)

        return wrapper

    return decorate



__version__ = '0.1.0'


import nv.tables.implementations.csv as csv
import nv.tables.implementations.xls as xls
import nv.tables.implementations.xlsx as xlsx

from nv.tables.implementations import *
from nv.tables.implementations import __ALL__ as __ALL_IMPLEMENTATIONS


__ALL__ = __ALL_IMPLEMENTATIONS
